<?php
/* vim: set expandtab tabstop=4 shiftwidth=4: */
//
// +----------------------------------------------------------------------+
// | PHP Version 4                                                        |
// +----------------------------------------------------------------------+
// | Copyright (c) 1997-2002 The PHP Group                                |
// +----------------------------------------------------------------------+
// | This source file is subject to version 2.02 of the PHP license,      |
// | that is bundled with this package in the file LICENSE, and is        |
// | available at through the world-wide-web at                           |
// | http://www.php.net/license/2_02.txt.                                 |
// | If you did not receive a copy of the PHP license and are unable to   |
// | obtain it through the world-wide-web, please send a note to          |
// | license@php.net so we can mail you a copy immediately.               |
// +----------------------------------------------------------------------+
// | Authors: Alexander Zhukov <alex@veresk.ru> Original port from Python |
// | Authors: Harry Fuecks <hfuecks@phppatterns.com> Port to PEAR + more  |
// +----------------------------------------------------------------------+
//
// $Id$
//
require_once('PEAR.php');
require_once('XML/HTMLSax/Attributes_Parser.php');
require_once('XML/HTMLSax/State_Machine.php');
/**
 * Handles parsing of HTML (and other badly formed XML)
 *
 * Example
 * <pre>
 * class MyHandler {
 *     function myOpenHandler(&$parser,$name,$attrs){};
 *     function myCloseHandler(&$parser,$name){};
 *     function myDataHandler(&$parser,$data){};
 *     function myPIHandler(&$parser,$target,$data){};
 *     function myEscapeHandler(&$parser,$data){};
 *     function myJaspHandler(&$parser,$data){};
 * }
 * $handler =& new MyHandler();
 * $parser = & new XML_HTMLSax();
 * $parser->set_object($handler);
 * $parser->set_element_handler('myOpenHandler','myCloseHandler');
 * $parser->set_data_handler('myDataHandler');
 * $parser->set_pi_handler('myPIHandler');
 * $parser->set_escape_handler('myEscapeHandler');
 * $parser->set_jasp_handler('myJaspHandler');
 * $doc = file_get_contents('http://www.php.net');
 * $parser->parse($doc);
 * </pre>
 * 
 * @author Alexander Zhukov <alex@veresk.ru>
 * @author Harry Fuecks <hfuecks@phppatterns.com>
 * @access public
 * @package XML_HTMLSax
 */

class XML_HTMLSax extends PEAR
{
    /**
     * Instance of State_Machine
     * @var object
     * @see State_Machine
     * @access private
     */
    var $sm;

    /**
     * Instance of user defined SAX handler
     * @var object
     * @access private
     */
    var $handler_object=false;

    /**
     * Class method opening tag listener
     * @var string
     * @access private
     */
    var $open_tag_handler=false;

    /**
     * Class method closing tag listener
     * @var string
     * @access private
     */
    var $close_tag_handler=false;

    /**
     * Class method character data listener
     * @var string
     * @access private
     */
    var $data_handler=false;

    /**
     * Class method for PI listener
     * @var string
     * @access private
     */
    var $pi_handler=false;

    /**
     * Class method for escape listener
     * @var string
     * @access private
     */
    var $escape_handler=false;

    /**
     * Class method for jasp listener
     * @var string
     * @access private
     */
    var $jasp_handler=false;

    /**
     * HTML document to parse
     * @var string
     * @access private
     */
    var $_html = '';

    /**
     * Temporary tag store
     * @var string
     * @access private
     */
    var $_tmp_tag = '';

    /**
     * Temporary text node store
     * @var string
     * @access private
     */
    var $_tmp_textNode = '';

    /**
     * Array of parser options
     * @var array
     * @access private
     */
    var $parser_options;

    /**
     * Constructs XML_HTMLSax
     * When using with Sax filters, make sure to instantiate by reference
     *
     * Example:
     * <pre>$parser= & new XML_HTMLSax();</pre>
     */
    function XML_HTMLSax()
    {
        $this->sm =& new State_Machine();
        $this->sm->text_node_start=true;
        $this->parser_options = array();
        $this->parser_options['skipWhiteSpace'] = 0;
        $this->parser_options['trimDataNodes'] = 0;
        $this->parser_options['breakOnLineFeed'] = 0;
        $this->parser_options['caseFolding'] = 0;
    }

    /**
     * Sets a parser option. Available options are;
     * <ul>
     * <li>skipWhiteSpace: skip white space</li>
     * <li>trimDataNodes: trim the character data inside an element</li>
     * <li>breakOnLineFeed: linefeeds in character data trigger data handler</li>
     * <li>caseFolding: Tag names converted to uppercase</li>
     * </ul>
     *
     * Example:
     * <pre>$parser->set_option('caseFolding',1);</pre>
     *
     * @param string option name
     * @param int option value
     * @return mixed
     * @access public
     */
    function set_option ($name,$value)
    {
        $options=array(
            'skipWhiteSpace',
            'trimDataNodes',
            'breakOnLineFeed',
            'caseFolding'
        );
        if ( in_array($name,$options) ) {
            $this->parser_options[$name] = $value;
            return true;
        } else {
            return $this->raiseError('XML_HTMLSax::set_option('.$name.') illegal');
        }
    }

    /**
     * Set handler object where SAX listener functions are available
     * This method must be used before specifying any handler methods
     *
     * Example:
     * <pre>
     * $myHandler = & new MyHandler();
     * $parser->set_object($myHandler);
     * </pre>
     *
     * @param object opening tag handler (class method)
     * @return mixed
     * @access public
     */
    function set_object (&$object)
    {
        if ( is_object($object) ) {
            $this->handler_object=&$object;
            return true;
        } else {
            return $this->raiseError('XML_HTMLSax::set_object requires '.
                'an object instance');
        }
    }

    /**
     * Sets the element listeners for opening and closing tags
     * These much be methods defined in the class instance passed
     * to set_object
     *
     * Example:
     * <pre>
     * $parser->set_element_handler('myOpenHandler','myCloseHandler');
     * </pre>
     *
     * @param string opening tag handler (class method) <tag>
     * @param string closing tag handler (class method) </tag>
     * @return mixed
     * @access public
     */
    function set_element_handler ($open,$close)
    {
        if ( method_exists ( $this->handler_object,$open ) && 
                method_exists ( $this->handler_object,$close ) ) {
            $this->open_tag_handler=$open;
            $this->close_tag_handler=$close;
            return true;
        } else {
            return $this->raiseError('XML_HTMLSax::set_element_handler - open '.
                'or close method does not exist in set_object');
        }
    }

    /**
     * Sets the data listener for the contents of elements
     * This much be a method defined in the class instance passed
     * to set_object
     *
     * Example:
     * <pre>
     * $parser->set_data_handler('myDataHandler');
     * </pre>
     *
     * @param string cdata handler (class method)
     * @return mixed
     * @access public
     */
    function set_data_handler ($data_handler)
    {
        if ( method_exists ( $this->handler_object,$data_handler ) ) {
            $this->data_handler=$data_handler;
            return true;
        } else {
            return $this->raiseError('XML_HTMLSax::set_data_handler - '.
                $data_handler.' does not exist in set_object');
        }
    }

    /**
     * Sets the PI listener for processing instructions
     * This much be a method defined in the class instance passed
     * to set_object
     *
     * Example:
     * <pre>
     * $parser->set_pi_handler('myPIHandler');
     * </pre>
     *
     * @param string PI handler (class method)
     * @return mixed
     * @access public
     */
    function set_pi_handler ($pi_handler)
    {
        if ( method_exists ( $this->handler_object,$pi_handler ) ) {
            $this->pi_handler=$pi_handler;
            return true;
        } else {
            return $this->raiseError('XML_HTMLSax::set_pi_handler - '.
                $pi_handler.' does not exist in set_object');
        }
    }

    /**
     * Sets the escape listener.
     * This much be a method defined in the class instance passed
     * to set_object
     *
     * Example:
     * <pre>
     * $parser->set_escape_handler('myEscapeHandler');
     * </pre>
     *
     * @param string escape handler (class method) <! >
     * @return mixed
     * @access public
     */
    function set_escape_handler ($escape_handler)
    {
        if ( method_exists ( $this->handler_object,$escape_handler ) ) {
            $this->escape_handler=$escape_handler;
            return true;
        } else {
            return $this->raiseError('XML_HTMLSax::escape_handler - '.
                $escape_handler.' does not exist in set_object');
        }
    }

    /**
     * Sets the JASP listener (JSP / ASP markup) e.g. <pre><% %></pre>
     * This much be a method defined in the class instance passed
     * to set_object
     *
     * Example:
     * <pre>
     * $parser->set_jasp_handler('myJaspHandler');
     * </pre>
     *
     * @param string JASP handler (class method) for <% %>
     * @return mixed
     * @access public
     */
    function set_jasp_handler ($jasp_handler)
    {
        if ( method_exists ( $this->handler_object,$jasp_handler ) ) {
            $this->jasp_handler=$jasp_handler;
            return true;
        } else {
            return $this->raiseError('XML_HTMLSax::jasp_handler - '.
                $jasp_handler.' does not exist in set_object');
        }
    }

    /**
     * Invokes parsing of HTML
     *
     * Example:
     * <pre>
     * $doc = file_get_contents('http://www.php.net');
     * $parser->parse($doc);
     * </pre>
     *
     * @param string contain HTML / XML document
     * @return void
     * @access public
     */
    function parse($html)
    {
        
        $this->_html = $html;
        $this->_parse();
    }

    /**
     * Performs the parsing operation
     *
     * @return void
     * @access private
     */
    function _parse()
    {
        for ($i = 0; $i < strlen($this->_html); $i++)
        {
            $chr = $this->_html[$i];
            if ( !$this->sm->ignore_entities && $chr == '<' ) 
            {
                $this->sm->left_b = true;
                $this->sm->text_node_start = false;
                $this->sm->text_node_end = true;
                switch ( $this->_html[$i+1] )
                {
                    case '?':
                        $this->sm->owned = true;
                        $this->sm->pi = true;
                        $this->sm->ignore_entities = true;
                    break;
                    case '%':
                        $this->sm->owned = true;
                        $this->sm->jasp = true;
                        $this->sm->ignore_entities = true;
                    break;
                    case '!':
                        $this->sm->owned = true;
                        $this->sm->escape = true;
                        if ($this->_html[$i+2].$this->_html[$i+3] == '--')
                        {
                            $this->sm->ignore_entities = true;
                        }
                    break;
                }
            }
            if ( $chr == '>') 
            {
                if ( !$this->sm->ignore_entities )
                {
                    $this->sm->right_b = true;
                }
                else if ( $this->sm->pi )
                {
                    if ( $this->_html[$i-1] == '?' )
                    {
                        $this->sm->ignore_entities = false;
                        $this->sm->right_b = true;
                    }
                }
                else if ( $this->sm->jasp )
                {
                    if ( $this->_html[$i-1] == '%' )
                    {
                        $this->sm->ignore_entities = false;
                        $this->sm->right_b = true;
                    }
                }
                else if ( $this->sm->escape )
                {
                    if ( $this->sm->ignore_entities )
                    {
                        if (substr($this->_html,$i-2,3) == '-->')
                        {
                            $this->sm->ignore_entities = false;
                            $this->sm->right_b = true;
                        }
                    }
                }
            }
            if ($chr == "\n" && $this->parser_options['breakOnLineFeed'] == 1 )
            {
                $this->sm->text_node_end = true;
            }
            $this->_doChar($chr);
        }
    }

    /**
     * Deals with parsing individual characters, triggering the
     * character data handler when the end of a data block is found
     *
     * @param string character to parse
     * @return void
     * @access private
     */
    function _doChar($chr)
    {
        if ($this->sm->text_node_start)
        {
                $this->_tmp_textNode .= $chr;
        }
        
        if ($this->sm->text_node_end)
        {
            if ($this->parser_options['trimDataNodes'] == 1)
            {
                $this->_tmp_textNode = trim($this->_tmp_textNode);
            }
            if ($this->parser_options['skipWhiteSpace'] == 1)
            {
                if (trim($this->_tmp_textNode) != '') {
                    $this->handle_data($this->_tmp_textNode);
                }
            }
            else
            {
                $this->handle_data($this->_tmp_textNode);
            }

            $this->_tmp_textNode = '';

            $this->sm->text_node_end = false;
        }
        
        if ($this->sm->left_b)
        {
            $this->_tmp_tag .= $chr;
        }

        if ($this->sm->right_b)
        {
            $this->_tagHandler(); 
            $this->sm->text_node_start = true;
        }
    }

    /**
     * Deals with element opening and closing tags, triggering
     * the element handlers as well as ignoring comments and
     * DocType declarations
     *
     * @return void
     * @access private
     */
    function _tagHandler()
    {
        $tag = substr(substr($this->_tmp_tag,1),0,-1);

        if ( !$this->sm->owned )
        {
            if (substr($tag,0,1) == '/')
            {
                $this->handle_end_tag(substr($tag,1));
            }
            else
            {
                $pos = strpos($tag,' ');
                if ($pos === false)
                {
                    $this->handle_start_tag($tag,Array());
                }
                else
                {
                    $a = new Attributes_Parser();
                    $attr = $a->parse(substr($tag,$pos));
                    $this->handle_start_tag(substr($tag,0,$pos),$attr);
                }
            }
        }
        else if ( $this->sm->pi )
        {
            $tag = substr(substr($tag,1),0,-1);
            $posSpace = strpos($tag,' ');
            $posFeed = strpos($tag,"\n");
            if ( $posFeed != false && $posFeed < $posSpace )
                $pos = $posFeed;
            else
                $pos = $posSpace;
            $this->handle_pi(substr($tag,0,$pos),substr($tag,$pos));
            $this->sm->owned = false;
            $this->sm->pi = false;
        }
        else if ( $this->sm->escape )
        {
            $tag = substr($tag,1);
            $this->handle_escape($tag);
            $this->sm->owned = false;
            $this->sm->escape = false;
        }
        else if ( $this->sm->jasp )
        {
            $tag = substr(substr($tag,1),0,-1);
            $this->handle_jasp($tag);
            $this->sm->owned = false;
            $this->sm->jasp = false;
        }
        $this->_tmp_tag = '';
        $this->sm->left_b = false;
        $this->sm->right_b = false;

    }

    /**
     * Triggers the opening tag handler
     *
     * @return void
     * @access private
     */
    function handle_start_tag($txt,$attr)
    {
        if ( $this->parser_options['caseFolding'] == 1 )
            $txt = strtoupper($txt);
        if ( $this->open_tag_handler && is_object ( $this->handler_object ) )
        {
            $this->handler_object->{$this->open_tag_handler}($this,$txt,$attr);
        }
    }

    /**
     * Triggers the data handler
     *
     * @return void
     * @access private
     */
    function handle_data($txt)
    {
        if ( $this->data_handler && is_object ( $this->handler_object ) )
        {
            $this->handler_object->{$this->data_handler}($this,$txt);
        }
    }

    /**
     * Triggers the closing tag handler
     *
     * @return void
     * @access private
     */
    function handle_end_tag($txt)
    {
        if ( $this->parser_options['caseFolding'] == 1 )
            $txt = strtoupper($txt);
        if ( $this->close_tag_handler && is_object ( $this->handler_object ) )
        {
            $this->handler_object->{$this->close_tag_handler}($this,$txt);
        }
    }

    /**
     * Triggers PI handler
     *
     * @return void
     * @access private
     */
    function handle_pi($target,$data)
    {
        if ( $this->pi_handler && is_object ( $this->handler_object ) )
        {
            $this->handler_object->{$this->pi_handler}($this,$target,$data);
        }
    }

    /**
     * Triggers the escape handler for <! > blocks
     *
     * @return void
     * @access private
     */
    function handle_escape($txt)
    {
        if ( $this->escape_handler && is_object ( $this->handler_object ) )
        {
            $this->handler_object->{$this->escape_handler}($this,$txt);
        }
    }

    /**
     * Triggers JASP handler for <% %>
     *
     * @return void
     * @access private
     */
    function handle_jasp($txt)
    {
        if ( $this->jasp_handler && is_object ( $this->handler_object ) )
        {
            $this->handler_object->{$this->jasp_handler}($this,$txt);
        }
    }
}
?>